/*
 * spi.c
 *
 *  Created on: 21 Aug 2011
 *      Author: Mike
 */

#include <usb.h>
#include <usb_com.h>
#include <radio_queue.h>
#include <stdio.h>
#include "spi.h"

#define CR '\r'
#define LF '\n'
#define null '\0'

#define COMMAND_0	(0x40 + 0)	// GO_IDLE_STATE
#define COMMAND_1	(0x40 + 1)	// SEND_OP_COND (MMC)
#define COMMAND_8	(0x40 + 8)	// SEND_IF_COND
#define COMMAND_16 	(0x40 + 16)	// SET BLOCKLEN
#define COMMAND_17 	(0x40 + 17)	// READ_SINGLE_BLOCK
#define COMMAND_24 	(0x40 + 24)	// WRITE_BLOCK
#define COMMAND_55 	(0x40 + 55)	// APP_CMD
#define COMMAND_58	(0x40 + 58)	// READ_OCR
#define ACOMMAND_41	(0xC0 + 41)	// SEND_OP_COND (SDC)

#define MAX_TRIES 20
#define BUFFER_SIZE 512
#define BLOCK_SIZE_BITS 9

#define gVersionString "v1.01 23rd August 2011\n(c) 2011 Mike McPherson"

void doCommands(void);
void doDeselect(uint8 usartID);
void doIdle(void);
void doInit(uint8 usartID);
void doReadData(uint8 usartID, uint32 block);
void doSelect(uint8 usartID);
void doWriteData(uint32 block);
char getChar(void);
int32 getNumber(char *terminator);
void main(void);
void putchar(char c);
void ShowBuffer( uint8 *buffer, uint16 size);
void showVersion(void);
char skipChars(char toSkip);
void updateLeds(void);
uint8 waitForResponse(uint8 usartID, uint8 response, uint32 timeoutMs);

uint32 block;

void doCommands(void)
{
	char charCom;
	int16 i;
	uint8 XDATA buffer[512];

	charCom = getChar();
	while (charCom != null)
	{
		switch (charCom)
		{
		case '?':
			// identify self
			//==============
			showVersion();
			break;
		case 'b':
			for (i=0; i<512; i++)
				buffer[i] = (i*i)&0xFF;
			doSelect(spiUsart0);
			for (i = 1; i<1000; i++)
				spiReadData(spiUsart0, buffer, 512);
			doDeselect(spiUsart0);
			break;
		case 'd':
			printf("deSelect\n");
			doDeselect(spiUsart0);
			break;
		case 'i':
			doInit(spiUsart0);
			break;
		case 'r':
			charCom = ' ';
			i = getNumber(&charCom);
			if (i>=0)
				block = i;
			printf("read block %ld\n", block);
			doSelect(spiUsart0);
			doReadData(spiUsart0, block);
			doDeselect(spiUsart0);
			block++;
			break;
		case 's':
			printf("select\n");
			doSelect(spiUsart0);
			break;
		case 'w':
			charCom = ' ';
			i = getNumber(&charCom);
			if (i<0)
				i = 0;
			printf("write block %d\n", i);
			doWriteData(i);
			break;
		case '\r':
			break;
		case '\n':
			break;
		default:
			printf("%c?\n", charCom);
			break;
		}
		charCom = getChar();
	}
}

void doDeselect(uint8 usartID)
{
	uint8 i;

	// deselect SPI device
	//====================
	i = spiDeselect(usartID);
	// printf("spiDeselect returned %x\n", i);
}

void doIdle(void)
{
	// service stuff we have to do to keep the world spinning
	//=======================================================
	boardService();
	updateLeds();
	usbComService();
}

void doInit(uint8 usartID)
{
	// initialise SD card
	//===================

	uint16 i, j;
	uint32 speed;
	uint8 XDATA buffer[BUFFER_SIZE];
	uint8 reply;

	// initialise SPI
	// USART0, alternate1, polarity negative, phase negative, MSB first
	//=================================================================
	printf("\ndoInit: initialise\n");
	i = spiInit(usartID, spiAlternate1, spiPolarityN, spiPhaseN, spiBitOrderM);
	printf("spiInit returned %d\n\n", i);

	// set SPI speed to 400 KHz initially, needed for SD initialisation
	//=================================================================
	printf("doInit: set 400 KHz\n");
	speed = spiSetSpeed(usartID, 400000);
	printf("spiSetSpeed returned %ld\n\n", speed);

	// deselect the device
	//====================
	doDeselect(usartID);

	// select the device
	//==================
	doSelect(usartID);

	// wait > 74 clock cycles
	// need to send dummy data so clock actually starts!
	// 10 bytes = 80 clocks
	//==================================================
	printf("doInit: clock[74]\n\n");
	for (i=0; i<10; i++)
		buffer[i] = 0xFF;
	spiWriteData(usartID, buffer, 10);

	reply = 0;
	for (i=0; i<10; i++)
	{
		// set card idle, with parameter 0 and checksum
		//=============================================
		printf("doInit: COMMAND_0\n");
		buffer[0] = COMMAND_0;
		buffer[1] = 0x00;
		buffer[2] = 0x00;
		buffer[3] = 0x00;
		buffer[4] = 0x00;
		buffer[5] = 0x95;	// fixed checksum
		spiWriteData(usartID, buffer, 6);

		// get reply, wait up to 1 second
		// looking for 0x01 = "gone idle"
		//===============================
		printf("doInit: wait for idle\n");
		reply = waitForResponse(usartID, 0x01, 1000);
		printf("waitForResponse returned %2X\n\n", reply);
		if (reply==1)
		{
			// success!
			break;
		}
	}

	// padding byte for clocks
	//========================
	spiWriteByte(usartID, 0xFF);

	// deselect the device
	//====================
	doDeselect(usartID);

	if (reply==0)
	{
		// didn't get good reply
		//======================
		printf("doinit: timeout waiting for idle\n");
		return;
	}

	// select the device
	//==================
	doSelect(usartID);

	// loop reading OCR
	// until either maximum attempts
	// or we get a good reply
	//==============================
	printf("doInit: READ_OCR\n");
	reply = 0;
	for (j=0; j<MAX_TRIES; j++)
	{
		// send "READ_OCR"
		//================
		buffer[0] = COMMAND_1;
		buffer[1] = 0;
		buffer[2] = 0;
		buffer[3] = 0;
		buffer[4] = 0;
		buffer[5] = 0xFF;	// dummy checksum
		buffer[6] = 0xFF;	// trailing padding
		spiWriteData(usartID, buffer, 6);

		spiWriteByte(usartID, 0xFF);

		// get reply
		//==========
		reply = spiReadByte(usartID);
		printf("spiReadByte returned %2X\n", reply);
		if (reply == 0x01)
		{
			// done initialising
			//==================
			break;
		}
	}

	if (reply == 0)
	{
		// we tried too many times
		//========================
		printf("doInit READ_OCR timed out after %d tries\n", MAX_TRIES);

		// deselect the device
		//====================
		doDeselect(usartID);

		return;
	}

	printf("\n");

	// send a few idle clocks
	// don't care what gets read
	//==========================
	printf("doInit: pause\n\n");
	spiReadByte(usartID);

	// set block size 512 bytes (MSB first)
	//=====================================
	printf("doInit: set block size\n");
	buffer[0] = COMMAND_16;
	buffer[1] = 0x00;
	buffer[2] = 0x00;
	buffer[3] = 0x02;
	buffer[4] = 0x00;
	buffer[5] = 0xFF;	// dummy checksum
	buffer[6] = 0xFF;	// trailing padding
	spiWriteData(usartID, buffer, 7);

	for (i=0; i<10; i++)
	{
		// get reply
		//==========
		printf("doInit: reply to set block size\n");
		reply = spiReadByte(usartID);
		printf("spiReadByte returned %2X\n\n", reply);
		if (reply == 0x00)
		{
			// good answer!
			//=============
			break;
		}
	}

	// padding byte for clocks
	//========================
	spiWriteByte(usartID, 0xFF);

	// deselect the device
	//====================
	doDeselect(usartID);

	// set SPI speed to 3 MHz (maximum)
	//=================================
	printf("doInit: set 3 MHz\n");
	speed = spiSetSpeed(usartID, 3000000);
	printf("spiSetSpeed returned %ld\n\n", speed);

	printf("doInit complete\n\n");
}

void doReadData(uint8 usartID, uint32 block)
{
	uint16 i;
	uint8 XDATA buffer[BUFFER_SIZE+2];	// add 2 bytes for checksum!
	uint16 size;
	// uint8 reply;

	// read data from block specified
	//===============================

	// empty the buffer
	//=================
	for (i=0; i<BUFFER_SIZE; i++)
		buffer[i] = 0;

	// convert block to address
	// block is 512 bytes (2^9)
	//===============================
	block <<= BLOCK_SIZE_BITS;

	// padding byte for clocks
	//========================
	spiWriteByte(usartID, 0xFF);

	// read block command
	//==================
	buffer[0] = COMMAND_17;
	buffer[1] = (block >> 24) & 0xFF;
	buffer[2] = (block >> 16) & 0xFF;
	buffer[3] = (block >>  8) & 0xFF;
	buffer[4] = (block      ) & 0xFF;
	buffer[5] = 0xFF;	// dummy checksum
	i = spiWriteData(usartID, buffer, 6);

	if (waitForResponse(usartID, 0xFE, 1000)==0)
	{
		// timed out :(
		//=============
		printf("timeout\n");
		return;
	}

/*
	printf("doReadData: wait for reply\n");
	reply = 0;
	for (i=0; i<25; i++)
	{
		// get reply
		//==========
		reply = spiReadByte(usartID);
		printf("spiReadByte returned %2X\n", reply);

		if (reply == 0xFE)
		{
			// good answer!
			//=============
			break;
		}
	}

	if (reply != 0xFE)
	{
		// didn't get data token
		//======================
		printf("no data\n");
		return;
	}
*/

	// read block of data
	//===================
	size = spiReadData(usartID, buffer, BUFFER_SIZE+2);
	printf("doReadData (read data) returned %d\n", size);

	// show the results
	//=================
	ShowBuffer(buffer, BUFFER_SIZE);

	printf("\n");
}

void doSelect(uint8 usartID)
{
	uint8 i;

	// select SPI device
	//==================
	i = spiSelect(usartID);
	// printf("spiSelect returned %x\n", i);
}

void doWriteData(uint32 block)
{
	uint16 i;
	uint8 XDATA buffer[BUFFER_SIZE];
	uint16 size;

	// write data to block specified
	// block is 512 bytes (2^9)
	//===============================
	block <<= BLOCK_SIZE_BITS;

	// write block command
	//====================
	buffer[0] = COMMAND_24;
	buffer[1] = (block >> 24) & 0xFF;
	buffer[2] = (block >> 16) & 0xFF;
	buffer[3] = (block >>  8) & 0xFF;
	buffer[4] = (block      ) & 0xFF;
	buffer[5] = 0xFF;	// dummy checksum
	buffer[6] = 0xFF;	// trailing padding

	i = spiWriteData(spiUsart0, buffer, 7);
	printf("spiWriteData (write command) returned %x\n", i);

	// initialise the buffer
	//====================
	for (i=0; i<BUFFER_SIZE; i++)
		buffer[i] = i & 0xFF;

	// write block of data
	//===================
	size = spiWriteData(spiUsart0, buffer, BUFFER_SIZE);
	printf("spiWritedData (data) returned %d\n", size);
}

char getChar(void)
{
	// function to get the next char from the COM port
	// returns null if nothing in yet
	//================================================

	uint8 charCount;
	uint8 charCom;

	charCount = usbComRxAvailable();

	if (charCount > 0)
	{
		charCom = usbComRxReceiveByte();
		return charCom;
	}
	else
	{
		return(null);
	}
}

int32 getNumber(char *terminator)
{
	char charCom;
	int32 P1;

	// read a number from COM
	// skips leading spaces
	// returns character that terminated the number
	// uses the terminator passed as the first possible character
	// returns -1 by default
	//===========================================================
	P1 = -1;

	// start with character passed
	//============================
	charCom = *terminator;

	// skip leading spaces
	//====================
	if (charCom==' ')
		charCom = skipChars(' ');

	// end of command?
	//================
	if (charCom==CR || charCom==LF)
	{
		// yup, so nothing to do!
		//=======================
		*terminator = charCom;
		return P1;
	}

	// read a P1 as a number
	//======================
	P1 = 0;
	while (charCom!=null && charCom>='0' && charCom<='9')
	{
		P1 = P1 * 10 + charCom - '0';
		charCom = getChar();
	}
	*terminator = charCom;
	return P1;
}

void main(void)
{
	// initialise lots of stuff
	//=========================
	systemInit();
	usbInit();
	radioQueueInit();

	block = 0;

	while (1)
	{
		doIdle();

		// now what we wanted to do!
		//==========================
		doCommands();
	}
}

void putchar(char c)
{
	// This gets called by puts, printf, and printBar
	//===============================================

	// wait for transmit ok
	//=====================
	if (usbComTxAvailable()>0)
	{
		// send a character
		//=================
		usbComTxSendByte(c);
	}
}

void ShowBuffer( uint8 *buffer, uint16 size)
{
	uint16 i, r, c, x;

	// print the buffer block
	//=======================
	r = 0;
	c = 0;
	do
	{
		if (size-r*16 < 16)
			i = size-r*16;
		else
			i = 16;

		printf("%04X: ", r*16);
		for (c=0; c<i; c++)
			printf("%02X.", buffer[r*16+c]);
		printf("  ");
		for (c=0; c<i; c++)
		{
			x = buffer[r*16+c];
			if (x<32 || x>126)
				x = '.';
			printf("%c", x);
		}
		printf("\n");
		r++;
	} while (r*16 < size);
}

void showVersion(void)
{
    printf("\nspi test Interface\n%s\nwixel: %2X-%2X-%2X-%2X\n\n",
    		gVersionString, serialNumber[3], serialNumber[2], serialNumber[1], serialNumber[0]);
}

char skipChars(char toSkip)
{
	char charCom;

	// wait for non-space character from COM port
	// returns 1st non-space character
	//===========================================
	charCom = toSkip;
	while ((charCom == toSkip) || (charCom == null))
	{
		charCom = getChar();
	}
	return charCom;
}

void updateLeds(void)
{
	// housekeeping routine
	//=====================
    usbShowStatusWithGreenLed();
}

uint8 waitForResponse(uint8 usartID, uint8 response, uint32 timeoutMs)
{
	uint32 startTime, timeSpent;
	uint8 reply;

	// keep reading data bytes until we get "response"
	// or timeout
	//
	// returns 1 on success
	// returns 0 on timeout
	//================================================

	startTime = getMs();

	while (1)
	{
		// keep board alive
		//=================
		doIdle();

		reply = spiReadByte(usartID);
		if (reply == response)
		{
			// got what we wanted
			//===================
			return 1;
		}
		else
		{
			if (reply != 0xFF)
				printf("waitForResponse expected %2X, got %2X\n", response, reply);
			timeSpent = getMs() - startTime;
			if (timeSpent > timeoutMs)
			{
				// too slow
				//=========
				break;
			}
		}
	}
	return 0;
}
